"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const moment = require("moment");
class BaseNode {
}
exports.BaseNode = BaseNode;
class NullNode extends BaseNode {
    solve() {
        throw new Error("AST node solved to null");
    }
}
exports.NullNode = NullNode;
class NumberNode extends BaseNode {
    constructor(value) {
        super();
        this.value = value;
    }
    solve() {
        return this.value;
    }
}
exports.NumberNode = NumberNode;
class StringNode extends BaseNode {
    constructor(value) {
        super();
        this.value = value;
    }
    solve() {
        return this.value;
    }
}
exports.StringNode = StringNode;
class TimestampNode extends NumberNode {
    constructor(value) {
        super(value);
        this.value = moment.unix(value).unix();
    }
}
exports.TimestampNode = TimestampNode;
class RelativeTokenNode extends BaseNode {
    constructor(token, quantity) {
        super();
        this.token = token;
        this.quantity = quantity;
    }
    solve() {
        return [
            this.quantity,
            this.token
        ];
    }
}
exports.RelativeTokenNode = RelativeTokenNode;
class RelativeExpressionNode extends BaseNode {
    constructor(anchor, relative) {
        super();
        this.anchor = anchor;
        this.relative = relative;
    }
    solve() {
        return this.calculateTimestamp().solve();
    }
    calculateTimestamp() {
        var timestamp = this.anchor.solve();
        // If a relative offset is provided, adjust for it
        if (this.relative && this.relative.token) {
            var ref = moment.unix(timestamp).utc();
            switch (this.relative.token) {
                case "M":
                    var rel = ref.clone().add(this.relative.quantity, "months").utc();
                    break;
                case "Q":
                    var rel = ref.clone().add(this.relative.quantity, "quarters").utc();
                    break;
                case "Y":
                    var rel = ref.clone().add(this.relative.quantity, "years").utc();
                    break;
                default:
                    throw new Error(`Relative token type ${this.relative.token} is not recognized`);
            }
            rel = this.correctMonthShift(ref, rel, this.relative.quantity);
            timestamp = rel.unix();
        }
        return new TimestampNode(timestamp);
    }
    /**
     * If adding 'period' time periods caused us to land on a date that does not exist (ex: Feb 30) moment will inconsitantly
     * shift us one way or anouther to a real date. We want to maintain a consistant behavoir, so we always enforce
     * that we shift conservatively (back when moving forwards, forward when moving backwards)
     *
     * This method detects if such a shift has occurred and returns a new moment that is correctly shifted
     *
     * @param  ref    The reference time before the translation
     * @param  rel    The computed time after the translation
     * @param  period The translation period
     * @return The time after the translation, adjusted for any month shift
     */
    correctMonthShift(ref, rel, period) {
        //
        var adjRelMoment = rel.clone();
        if (ref.date() != rel.date()) {
            if (period >= 0 && rel.date() != rel.clone().endOf('month').date()) {
                adjRelMoment = rel.clone().subtract(1, 'day');
            }
            else if (period < 0 && rel.date() != rel.clone().startOf('month').date()) {
                adjRelMoment = rel.clone().add(1, 'day');
            }
        }
        return adjRelMoment;
    }
}
exports.RelativeExpressionNode = RelativeExpressionNode;
class FunctionNode extends BaseNode {
}
exports.FunctionNode = FunctionNode;
class RangeNode extends FunctionNode {
    constructor(start, end) {
        super();
        this.Start = start;
        this.End = end;
    }
    solve() {
        return [this.Start.solve(),
            this.End.solve()].sort();
    }
}
exports.RangeNode = RangeNode;
class TrailNode extends FunctionNode {
    constructor(end, period) {
        super();
        this.End = end;
        this.Period = period;
    }
    solve() {
        var end = this.End.solve();
        var start;
        switch (this.Period.constructor) {
            case RelativeTokenNode:
                // need to reconstruct the relative token to face the opposite direction (to trail rather than lead)
                var [quantity, token] = this.Period.solve();
                var periodTokenNode = new RelativeTokenNode(token, -1 * quantity);
                start = new RelativeExpressionNode(new TimestampNode(this.End.solve()), periodTokenNode).solve();
                break;
            case RelativeExpressionNode:
                start = end - this.Period.solve();
                break;
            default:
                throw new Error("Something went horribly wrong");
        }
        return [start, end].sort();
    }
}
exports.TrailNode = TrailNode;
class PeriodNode extends FunctionNode {
    constructor(start, period, ref) {
        super();
        this.Start = start;
        this.Period = period;
        this.Ref = ref;
    }
    solve() {
        var period;
        var offset;
        var start;
        var end;
        var anchor = this.Start.solve();
        var ref = this.Ref.solve();
        var diff = ref - anchor;
        switch (this.Period.constructor) {
            case RelativeTokenNode:
                // because relative tokens do not have fixed lengths (the length of a month/quarter/year fluxuates with context)
                // we first approximate the periods we are incrementing by to find out about how many times to shift by the relative
                // token. We then perform the shift and step up or down until we reach the correct period
                var [quantity, token] = this.Period.solve();
                var firstBoundary = new RelativeExpressionNode(new TimestampNode(anchor), this.Period).solve();
                var apprxOffset = Math.floor(diff / (firstBoundary - anchor));
                start = new RelativeExpressionNode(new TimestampNode(anchor), new RelativeTokenNode(token, quantity * apprxOffset)).solve();
                end = new RelativeExpressionNode(new TimestampNode(start), new RelativeTokenNode(token, quantity * Math.sign(diff))).solve();
                var [start, end] = this.relShiftToCorrectPeriod(start, end, ref, token, quantity);
                break;
            case RelativeExpressionNode:
                period = this.Period.solve();
                offset = Math.floor(diff / period);
                var [start, end] = this.getPeriodBounds(anchor, period, offset, diff);
                var [start, end] = this.shiftToCorrectPeriod(start, end, ref, period);
                break;
            default:
                throw new Error("Something has gone horribly wrong");
        }
        return [start, end].sort();
    }
    getPeriodBounds(anchor, period, offset, diff) {
        var start;
        var end;
        switch (Math.sign(diff)) {
            case 1:
                start = anchor + offset * period;
                end = start + period;
                break;
            case 0:
                // Always go back if lying directly on a boundary
                start = anchor - period;
                end = anchor;
                break;
            case -1:
                end = anchor - offset * period;
                start = end - period;
                break;
            default:
                throw new Error("Something has gone horribly wrong");
        }
        return [start, end];
    }
    shiftToCorrectPeriod(start, end, ref, period) {
        while (!(start < ref && ref <= end)) {
            if (ref <= start) {
                start -= period;
                end -= period;
            }
            else if (ref > end) {
                start += period;
                end += period;
            }
        }
        return [start, end];
    }
    relShiftToCorrectPeriod(start, end, ref, token, quantity) {
        quantity = Math.abs(quantity);
        while (!(start < ref && ref <= end)) {
            if (ref <= start) {
                start = new RelativeExpressionNode(new TimestampNode(start), new RelativeTokenNode(token, -quantity)).solve();
                end = new RelativeExpressionNode(new TimestampNode(end), new RelativeTokenNode(token, -quantity)).solve();
            }
            else if (ref > end) {
                start = new RelativeExpressionNode(new TimestampNode(start), new RelativeTokenNode(token, quantity)).solve();
                end = new RelativeExpressionNode(new TimestampNode(end), new RelativeTokenNode(token, quantity)).solve();
            }
        }
        return [start, end];
    }
}
exports.PeriodNode = PeriodNode;
