"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("mocha");
const chai_1 = require("chai");
const helpers_1 = require("./helpers");
const ast_nodes_1 = require("../ast-nodes");
describe("FunctionFactory", function () {
    describe("Basic Functionality", function () {
        it("should allow custom defined 'now'", function () {
            let tree = helpers_1.getParseTreeFromString("$range$0$now$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 150);
            chai_1.expect(func.End.anchor.solve()).to.be.equal(150);
        });
    });
    describe("Range Function", function () {
        it("should produce a RangeNode", function () {
            let tree = helpers_1.getParseTreeFromString("$range$0$now$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func).to.be.instanceOf(ast_nodes_1.RangeNode);
        });
        it("should have the correct start and end nodes", function () {
            let tree = helpers_1.getParseTreeFromString("$range$111$999$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func).to.be.eql(new ast_nodes_1.RangeNode(new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(111)), new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(999))));
        });
        it("should have the correct start and end nodes (now reference)", function () {
            let tree = helpers_1.getParseTreeFromString("$range$111$now$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 999);
            chai_1.expect(func).to.be.eql(new ast_nodes_1.RangeNode(new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(111)), new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(999))));
        });
        it("should have the correct start and end nodes (static expression)", function () {
            let tree = helpers_1.getParseTreeFromString("$range$111-1s$now+1s$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 999);
            chai_1.expect(func).to.be.eql(new ast_nodes_1.RangeNode(new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(110)), new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(1000))));
        });
        it("should have the correct start and end nodes (relative expression)", function () {
            let tree = helpers_1.getParseTreeFromString("$range$111-1Q$now+1Y$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 999);
            chai_1.expect(func).to.be.eql(new ast_nodes_1.RangeNode(new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(111), new ast_nodes_1.RelativeTokenNode('Q', -1)), new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(999), new ast_nodes_1.RelativeTokenNode('Y', 1))));
        });
        it("should solve to correct timestamp range (static)", function () {
            let tree = helpers_1.getParseTreeFromString("$range$1561872660$1561872880$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func.solve()).to.be.eql([1561872660, 1561872880]);
        });
    });
    describe("Trail Function", function () {
        it("should produce a TraiNode", function () {
            let tree = helpers_1.getParseTreeFromString("$trail$now$5m$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func).to.be.instanceOf(ast_nodes_1.TrailNode);
        });
        it("should have the correct end and period nodes", function () {
            let tree = helpers_1.getParseTreeFromString("$trail$111$5m$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func).to.be.eql(new ast_nodes_1.TrailNode(new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(111)), new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(5 * 60 * 1))));
        });
        it("should have the correct end and period nodes (relative)", function () {
            let tree = helpers_1.getParseTreeFromString("$trail$1561872660$1Q$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func).to.be.eql(new ast_nodes_1.TrailNode(new ast_nodes_1.RelativeExpressionNode(new ast_nodes_1.TimestampNode(1561872660)), new ast_nodes_1.RelativeTokenNode('Q', 1)));
            chai_1.expect(func.solve()).to.be.eql([1553923860, 1561872660]);
        });
        it("should lead if the provided trail is negative", function () {
            let tree = helpers_1.getParseTreeFromString("$trail$1553923860$-1Q$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func.solve()).to.be.eql([1553923860, 1561872660]);
        });
        it("should solve to correct timestamp range (static)", function () {
            let tree = helpers_1.getParseTreeFromString("$trail$1561872660$60$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func.solve()).to.be.eql([1561872600, 1561872660]);
        });
        it("should solve to correct timestamp range (relative)", function () {
            let tree = helpers_1.getParseTreeFromString("$trail$1561872660$1Q$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func.solve()).to.be.eql([1553923860, 1561872660]);
        });
    });
    describe("Period Function", function () {
        it("should produce a PeriodNode", function () {
            let tree = helpers_1.getParseTreeFromString("$period$now$5m$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func).to.be.instanceOf(ast_nodes_1.PeriodNode);
        });
        it("should produce a PeriodNode when given ref", function () {
            let tree = helpers_1.getParseTreeFromString("$period$now$5Q$now$");
            let func = helpers_1.getFunctionTreeFromDirective(tree);
            chai_1.expect(func).to.be.instanceOf(ast_nodes_1.PeriodNode);
        });
        it("should solve to the correct timestamp range (first +ve period, static)", function () {
            let tree = helpers_1.getParseTreeFromString("$period$1561872660$10s$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 1561872661);
            chai_1.expect(func.solve()).to.be.eql([1561872660, 1561872670]);
        });
        it("should solve to the correct timestamp range (first -ve period, static)", function () {
            let tree = helpers_1.getParseTreeFromString("$period$1561872660$10s$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 1561872659);
            chai_1.expect(func.solve()).to.be.eql([1561872650, 1561872660]);
        });
        it("should solve to the correct timestamp range (nth +ve period, static, on boundary)", function () {
            let tree = helpers_1.getParseTreeFromString("$period$1561872660$10s$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 1561872690);
            chai_1.expect(func.solve()).to.be.eql([1561872680, 1561872690]);
        });
        it("should solve to the correct timestamp range (first +ve period, relative)", function () {
            let tree = helpers_1.getParseTreeFromString("$period$1561872660$1Q$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 1561872800);
            chai_1.expect(func.solve()).to.be.eql([1561872660, 1569821460]);
        });
        it("should solve to the correct timestamp range (nth +ve period, relative)", function () {
            let tree = helpers_1.getParseTreeFromString("$period$1561872660$1Q$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 1569821461);
            chai_1.expect(func.solve()).to.be.eql([1569821460, 1577683860]);
        });
        it("should solve to the correct timestamp range (nth +ve period, relative, adjusted ref)", function () {
            let tree = helpers_1.getParseTreeFromString("$period$1561872660$1Q$now-5m$");
            let func = helpers_1.getFunctionTreeFromDirective(tree, 1569821461);
            chai_1.expect(func.solve()).to.be.eql([1561872660, 1569821460]);
        });
    });
    describe("Relative Expression Node", function () {
        it("should solve to a correct timestamp node when given no relative", function () {
            var anchor = new ast_nodes_1.TimestampNode(13579);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor);
            chai_1.expect(node.solve()).to.be.eql(13579);
        });
        it("should properly traverse 0 months", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', 0);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
        });
        it("should properly traverse 0 quarters", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Q', 0);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
        });
        it("should properly traverse 0 years", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Y', 0);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
        });
        it("should properly traverse back 1 month", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', -1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 3, 15, 5, 30) / 1000);
        });
        it("should properly traverse forward 1 month", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', 1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 5, 15, 5, 30) / 1000);
        });
        it("should properly traverse back n months (within year)", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', -4);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 0, 15, 5, 30) / 1000);
        });
        it("should properly traverse back n months (across year)", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', -6);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2018, 10, 15, 5, 30) / 1000);
        });
        it("should properly traverse back n months (across year && > 12)", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', -25);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2017, 3, 15, 5, 30) / 1000);
        });
        it("should properly traverse forward n months (within year)", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', 4);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 8, 15, 5, 30) / 1000);
        });
        it("should properly traverse forward n months (across year)", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', 10);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2020, 2, 15, 5, 30) / 1000);
        });
        it("should properly traverse forward n months (across year && > 12)", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', 25);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2021, 5, 15, 5, 30) / 1000);
        });
        it("(0) should be conservative when traversing back to a non-existant date by months", function () {
            var anchor_0 = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 2, 29, 5, 30) / 1000);
            var anchor_1 = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 2, 30, 5, 30) / 1000);
            var anchor_2 = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 2, 31, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', -1);
            var node_0 = new ast_nodes_1.RelativeExpressionNode(anchor_0, relative);
            var node_1 = new ast_nodes_1.RelativeExpressionNode(anchor_1, relative);
            var node_2 = new ast_nodes_1.RelativeExpressionNode(anchor_2, relative);
            chai_1.expect(node_0.solve()).to.be.eql(+Date.UTC(2019, 2, 1, 5, 30) / 1000);
            chai_1.expect(node_1.solve()).to.be.eql(+Date.UTC(2019, 2, 1, 5, 30) / 1000);
            chai_1.expect(node_2.solve()).to.be.eql(+Date.UTC(2019, 2, 1, 5, 30) / 1000);
        });
        it("(1) should be conservative when traversing back to a non-existant date by months", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 31, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', -1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 4, 1, 5, 30) / 1000);
        });
        it("(0) should be conservative when traversing forward to a non-existant date by months", function () {
            var anchor_0 = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 0, 29, 5, 30) / 1000);
            var anchor_1 = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 0, 30, 5, 30) / 1000);
            var anchor_2 = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 0, 31, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', 1);
            var node_0 = new ast_nodes_1.RelativeExpressionNode(anchor_0, relative);
            var node_1 = new ast_nodes_1.RelativeExpressionNode(anchor_1, relative);
            var node_2 = new ast_nodes_1.RelativeExpressionNode(anchor_2, relative);
            chai_1.expect(node_0.solve()).to.be.eql(+Date.UTC(2019, 1, 28, 5, 30) / 1000);
            chai_1.expect(node_1.solve()).to.be.eql(+Date.UTC(2019, 1, 28, 5, 30) / 1000);
            chai_1.expect(node_2.solve()).to.be.eql(+Date.UTC(2019, 1, 28, 5, 30) / 1000);
        });
        it("(1) should be conservative when traversing forward to a non-existant date by months", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 31, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('M', 1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 5, 30, 5, 30) / 1000);
        });
        it("should properly traverse back 1 year", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Y', -1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2018, 4, 15, 5, 30) / 1000);
        });
        it("should properly traverse forward 1 year", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Y', 1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2020, 4, 15, 5, 30) / 1000);
        });
        it("should properly traverse back n years", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Y', -4);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2015, 4, 15, 5, 30) / 1000);
        });
        it("should properly traverse forward n years", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Y', 4);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2023, 4, 15, 5, 30) / 1000);
        });
        it("(0) should be conservative when traversing back to a non-existant date by years", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2020, 1, 29, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Y', -1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 2, 1, 5, 30) / 1000);
        });
        it("(1) should be conservative when traversing forward to a non-existant date by years", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2020, 1, 29, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Y', 1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2021, 1, 28, 5, 30) / 1000);
        });
        it("should properly traverse back 1 quarter", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Q', -1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 1, 15, 5, 30) / 1000);
        });
        it("should properly traverse forward 1 quarter", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Q', 1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 7, 15, 5, 30) / 1000);
        });
        it("should properly traverse back n quarters", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Q', -3);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2018, 7, 15, 5, 30) / 1000);
        });
        it("should properly traverse forward n quarters", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 15, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Q', 3);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2020, 1, 15, 5, 30) / 1000);
        });
        it("(0) should be conservative when traversing back to a non-existant date by quarters", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2019, 4, 29, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Q', -1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 2, 1, 5, 30) / 1000);
        });
        it("(1) should be conservative when traversing forward to a non-existant date by quarters", function () {
            var anchor = new ast_nodes_1.TimestampNode(+Date.UTC(2018, 10, 29, 5, 30) / 1000);
            var relative = new ast_nodes_1.RelativeTokenNode('Q', 1);
            var node = new ast_nodes_1.RelativeExpressionNode(anchor, relative);
            chai_1.expect(node.solve()).to.be.eql(+Date.UTC(2019, 1, 28, 5, 30) / 1000);
        });
    });
});
